// distribution boxbackup-0.11_trunk_2979 (svn version: 2979)
// Box Backup, http://www.boxbackup.org/
// 
// Copyright (c) 2003-2010, Ben Summers and contributors.
// All rights reserved.
// 
// Note that this project uses mixed licensing. Any file with this license
// attached, or where the code LICENSE-DUAL appears on the first line, falls
// under this license. See the file COPYING.txt for more information.
// 
// This file is dual licensed. You may use and distribute it providing that you
// comply EITHER with the terms of the BSD license, OR the GPL license. It is
// not necessary to comply with both licenses, only one.
// 
// The BSD license option follows:
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  
// 3. Neither the name of the Box Backup nor the names of its contributors may
//    be used to endorse or promote products derived from this software without
//    specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// [http://en.wikipedia.org/wiki/BSD_licenses#3-clause_license_.28.22New_BSD_License.22.29]
// 
// The GPL license option follows:
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// [http://www.gnu.org/licenses/old-licenses/gpl-2.0.html#SEC4]
// --------------------------------------------------------------------------
//
// File
//		Name:    BufferedWriteStream.cpp
//		Purpose: Buffering write-only wrapper around IOStreams
//		Created: 2010/09/13
//
// --------------------------------------------------------------------------

#include "Box.h"
#include "BufferedWriteStream.h"
#include "CommonException.h"

#include <string.h>

#include "MemLeakFindOn.h"

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::BufferedWriteStream(const char *, int, int)
//		Purpose: Constructor, set up buffer
//		Created: 2007/01/16
//
// --------------------------------------------------------------------------
BufferedWriteStream::BufferedWriteStream(IOStream& rSink)
: mrSink(rSink), mBufferPosition(0)
{ }


// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::Read(void *, int)
//		Purpose: Reads bytes from the file - throws exception
//		Created: 2007/01/16
//
// --------------------------------------------------------------------------
int BufferedWriteStream::Read(void *pBuffer, int NBytes, int Timeout)
{
	THROW_EXCEPTION(CommonException, NotSupported);
}


// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::BytesLeftToRead()
//		Purpose: Returns number of bytes to read (may not be most efficient function ever)
//		Created: 2007/01/16
//
// --------------------------------------------------------------------------
IOStream::pos_type BufferedWriteStream::BytesLeftToRead()
{
	THROW_EXCEPTION(CommonException, NotSupported);
}


// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::Write(void *, int)
//		Purpose: Writes bytes to the underlying stream (not supported)
//		Created: 2003/07/31
//
// --------------------------------------------------------------------------
void BufferedWriteStream::Write(const void *pBuffer, int NBytes)
{
	int numBytesRemain = NBytes;

	do
	{
		int maxWritable = sizeof(mBuffer) - mBufferPosition;
		int numBytesToWrite = (numBytesRemain < maxWritable) ?
			numBytesRemain : maxWritable;

		if(numBytesToWrite > 0)
		{
			memcpy(mBuffer + mBufferPosition, pBuffer,
				numBytesToWrite);
			mBufferPosition += numBytesToWrite;
			pBuffer = ((const char *)pBuffer) + numBytesToWrite;
			numBytesRemain -= numBytesToWrite;
		}

		if(numBytesRemain > 0)
		{
			Flush();
		}
	}
	while(numBytesRemain > 0);
}

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::GetPosition()
//		Purpose: Get position in stream
//		Created: 2003/08/21
//
// --------------------------------------------------------------------------
IOStream::pos_type BufferedWriteStream::GetPosition() const
{
	return mrSink.GetPosition() + mBufferPosition;
}


// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::Seek(pos_type, int)
//		Purpose: Seeks within file, as lseek, invalidate buffer
//		Created: 2003/07/31
//
// --------------------------------------------------------------------------
void BufferedWriteStream::Seek(IOStream::pos_type Offset, int SeekType)
{
	// Always flush the buffer before seeking
	Flush();

	mrSink.Seek(Offset, SeekType);
}

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::Flush();
//		Purpose: Write out current buffer contents and invalidate
//		Created: 2010/09/13
//
// --------------------------------------------------------------------------
void BufferedWriteStream::Flush(int Timeout)
{
	if(mBufferPosition > 0)
	{
		mrSink.Write(mBuffer, mBufferPosition);
	}

	mBufferPosition = 0;
}

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::Close()
//		Purpose: Closes the underlying stream (not needed)
//		Created: 2003/07/31
//
// --------------------------------------------------------------------------
void BufferedWriteStream::Close()
{
	Flush();
	mrSink.Close();
}

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::StreamDataLeft()
//		Purpose: Any data left to write?
//		Created: 2003/08/02
//
// --------------------------------------------------------------------------
bool BufferedWriteStream::StreamDataLeft()
{
	THROW_EXCEPTION(CommonException, NotSupported);
}

// --------------------------------------------------------------------------
//
// Function
//		Name:    BufferedWriteStream::StreamClosed()
//		Purpose: Is the stream closed?
//		Created: 2003/08/02
//
// --------------------------------------------------------------------------
bool BufferedWriteStream::StreamClosed()
{
	return mrSink.StreamClosed();
}

