// distribution boxbackup-0.11_trunk_2979 (svn version: 2979)
// Box Backup, http://www.boxbackup.org/
// 
// Copyright (c) 2003-2010, Ben Summers and contributors.
// All rights reserved.
// 
// Note that this project uses mixed licensing. Any file with this license
// attached, or where the code LICENSE-DUAL appears on the first line, falls
// under this license. See the file COPYING.txt for more information.
// 
// This file is dual licensed. You may use and distribute it providing that you
// comply EITHER with the terms of the BSD license, OR the GPL license. It is
// not necessary to comply with both licenses, only one.
// 
// The BSD license option follows:
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
// 1. Redistributions of source code must retain the above copyright
//    notice, this list of conditions and the following disclaimer.
// 
// 2. Redistributions in binary form must reproduce the above copyright
//    notice, this list of conditions and the following disclaimer in the
//    documentation and/or other materials provided with the distribution.
//  
// 3. Neither the name of the Box Backup nor the names of its contributors may
//    be used to endorse or promote products derived from this software without
//    specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
// ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
// 
// [http://en.wikipedia.org/wiki/BSD_licenses#3-clause_license_.28.22New_BSD_License.22.29]
// 
// The GPL license option follows:
// 
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
// 
// [http://www.gnu.org/licenses/old-licenses/gpl-2.0.html#SEC4]
// --------------------------------------------------------------------------
//
// File
//		Name:    RateLimitingStream.cpp
//		Purpose: Rate-limiting write-only wrapper around IOStreams
//		Created: 2011/01/11
//
// --------------------------------------------------------------------------

#include "Box.h"
#include "RateLimitingStream.h"
#include "CommonException.h"

#include <string.h>

#include "MemLeakFindOn.h"

// --------------------------------------------------------------------------
//
// Function
//		Name:    RateLimitingStream::RateLimitingStream(const char *, int, int)
//		Purpose: Constructor, set up buffer
//		Created: 2011/01/11
//
// --------------------------------------------------------------------------
RateLimitingStream::RateLimitingStream(IOStream& rSink, size_t targetBytesPerSecond)
: mrSink(rSink), mStartTime(GetCurrentBoxTime()), mTotalBytesRead(0),
  mTargetBytesPerSecond(targetBytesPerSecond)
{ }

// --------------------------------------------------------------------------
//
// Function
//		Name:    RateLimitingStream::Read(void *pBuffer, int NBytes,
//			 int Timeout)
//		Purpose: Reads bytes to the underlying stream at no more than
//			 a fixed rate
//		Created: 2011/01/11
//
// --------------------------------------------------------------------------
int RateLimitingStream::Read(void *pBuffer, int NBytes, int Timeout)
{
	if(NBytes > 0 && (size_t)NBytes > mTargetBytesPerSecond)
	{
		// Limit to one second's worth of data for performance
		BOX_TRACE("Reducing read size from " << NBytes << " to " <<
			mTargetBytesPerSecond << " to smooth upload rate");
		NBytes = mTargetBytesPerSecond;
	}

	int bytesReadThisTime = mrSink.Read(pBuffer, NBytes, Timeout);

	// How many bytes we will have written after this write finishes?
	mTotalBytesRead += bytesReadThisTime;

	// When should it be completed by?
	box_time_t desiredFinishTime = mStartTime +
		SecondsToBoxTime(mTotalBytesRead / mTargetBytesPerSecond);

	// How long do we have to wait?
	box_time_t currentTime = GetCurrentBoxTime();
	int64_t waitTime = desiredFinishTime - currentTime;

	// How are we doing so far? (for logging only)
	box_time_t currentDuration = currentTime - mStartTime;

	// in case our timer is not very accurate, don't divide by zero on first pass
	if(currentDuration == 0)
	{
		BOX_TRACE("Current rate not yet known, sending immediately");
		return bytesReadThisTime;
	}
		
	uint64_t effectiveRateSoFar = (mTotalBytesRead * MICRO_SEC_IN_SEC_LL)
		/ currentDuration;

	if(waitTime > 0)
	{
		BOX_TRACE("Current rate " << effectiveRateSoFar <<
			" higher than desired rate " << mTargetBytesPerSecond <<
			", sleeping for " << BoxTimeToMilliSeconds(waitTime) <<
			" ms");
		ShortSleep(waitTime, false);
	}
	else
	{
		BOX_TRACE("Current rate " << effectiveRateSoFar <<
			" lower than desired rate " << mTargetBytesPerSecond <<
			", sending immediately (would have sent " <<
			(BoxTimeToMilliSeconds(-waitTime)) << " ms ago)");
	}

	return bytesReadThisTime;
}

